package org.processmining.models.dcnet;

import java.util.Set;

import org.processmining.datapetrinets.expression.GuardExpression;
import org.processmining.models.cnet.CNetBinding;
import org.processmining.models.cnet.CNetNode;
import org.processmining.models.cnet.CausalNet;

import com.google.common.collect.ImmutableSet;

/**
 * A data-aware causal net. In a data-aware causal net some output bindings may
 * be guarded {@link #getGuardedBindings()} by expressions
 * {@link #getGuardExpression(CNetBinding)} defined over variables
 * {@link #getVariables()}. Moreover, tasks write record variable values
 * {@link #getWrites(CNetNode)}.
 * <p>
 * To create a {@link DataAwareCausalNet} use, for example, the implementation
 * {@link DCNet}. The interface deliberately does not specify mutating methods
 * to be kept simple.
 * 
 * @author F. Mannhardt
 *
 */
public interface DataAwareCausalNet {

	public final class Factory {

		private static final class DataAwareCausalNetView implements DataAwareCausalNet {

			private final CausalNet causalNet;

			private DataAwareCausalNetView(CausalNet causalNet) {
				this.causalNet = causalNet;
			}

			public Set<CNetBinding> getGuardedBindings() {
				return ImmutableSet.of();
			}

			public GuardExpression getGuardExpression(CNetBinding binding) {
				return null;
			}

			public CausalNet getCausalNet() {
				return causalNet;
			}

			public Set<String> getVariables() {
				return ImmutableSet.of();
			}

			public Class<?> getVariableType(String variable) {
				return null;
			}

			public Set<String> getWrites(CNetNode node) {
				return ImmutableSet.of();
			}
		}

		private Factory() {
		}

		public static DataAwareCausalNet fromCausalNet(CausalNet causalNet) {
			return new DCNet(causalNet);
		}

		public static CausalNet toCausalNet(DataAwareCausalNet dataCausalNet) {
			return dataCausalNet.getCausalNet();
		}

		public static DataAwareCausalNet viewAsDataAwareCausalNet(final CausalNet causalNet) {
			return new DataAwareCausalNetView(causalNet);
		}

	}

	/**
	 * @return the underlying {@link CausalNet}
	 */
	CausalNet getCausalNet();

	/**
	 * @return the (possibly immutable) set of {@link CNetBinding} with a
	 *         {@link GuardExpression}
	 */
	Set<CNetBinding> getGuardedBindings();

	/**
	 * @param binding
	 * @return the {@link GuardExpression} if {@link CNetBinding} is in
	 *         {@link #getGuardedBindings()}, otherwise NULL
	 */
	GuardExpression getGuardExpression(CNetBinding binding);

	/**
	 * @return the (possibly immutable) set of variables used in the
	 *         {@link #getGuardedBindings()}
	 */
	Set<String> getVariables();

	/**
	 * @param variable
	 * @return the Java type of the variable, returns NULL when the variable is
	 *         unknown
	 */
	Class<?> getVariableType(String variable);

	/**
	 * @param task
	 * @return the set of variables supposed to be written by the
	 *         {@link CNetNode}
	 */
	Set<String> getWrites(CNetNode task);

}