package org.processmining.plugins.dcnet;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Point;
import java.awt.Shape;
import java.awt.Stroke;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.geom.AffineTransform;
import java.awt.geom.Ellipse2D;
import java.awt.geom.GeneralPath;
import java.awt.geom.NoninvertibleTransformException;
import java.awt.geom.PathIterator;
import java.awt.geom.Point2D;
import java.awt.geom.Point2D.Double;
import java.awt.geom.Rectangle2D;
import java.awt.geom.RectangularShape;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import javax.swing.SwingUtilities;

import org.processmining.models.cnet.CNetBinding;
import org.processmining.models.cnet.CNetNode;
import org.processmining.models.cnet.CausalNet;
import org.processmining.plugins.dcnet.BindingNode.DefaultBindingNodeSelectionListener;
import org.processmining.plugins.dcnet.BindingNode.SelectionListener;
import org.processmining.plugins.dcnet.NodeDecorator.BorderStyle;
import org.processmining.plugins.dcnet.NodeDecorator.ColorSegment;
import org.processmining.plugins.dcnet.NodeDecorator.TrafficLightSegment;
import org.processmining.plugins.graphviz.dot.Dot;
import org.processmining.plugins.graphviz.dot.DotEdge;
import org.processmining.plugins.graphviz.dot.DotElement;
import org.processmining.plugins.graphviz.dot.DotNode;
import org.processmining.plugins.graphviz.visualisation.DotPanel;
import org.processmining.plugins.graphviz.visualisation.export.Exporter;
import org.processmining.plugins.graphviz.visualisation.export.ExporterPDF;

import com.google.common.collect.BiMap;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Iterators;
import com.google.common.collect.Lists;
import com.google.common.collect.Ordering;
import com.google.common.collect.PeekingIterator;
import com.google.common.primitives.Ints;
import com.kitfox.svg.Group;
import com.kitfox.svg.Path;
import com.kitfox.svg.Polygon;
import com.kitfox.svg.SVGDiagram;
import com.kitfox.svg.SVGElement;
import com.kitfox.svg.SVGException;
import com.kitfox.svg.SVGRoot;
import com.kitfox.svg.Text;
import com.kitfox.svg.animation.Bezier;

final class CausalNetDotPanel extends DotPanel {

	static final class ConnectorImpl implements Connector {

		private final Point2D point;
		private final boolean isCorrectPosition;

		public ConnectorImpl(Point2D point) {
			this.point = point;
			this.isCorrectPosition = true;
		}

		public ConnectorImpl(Point2D point, boolean isCorrectPosition) {
			this.point = point;
			this.isCorrectPosition = isCorrectPosition;
		}

		public boolean isCorrectPosition() {
			return isCorrectPosition;
		}

		public Point2D getPosition() {
			return point;
		}

	}

	private static final long serialVersionUID = -7515672307835050920L;

	private final CausalNet causalNet;

	private final Set<BindingNode> bindingNodes = new HashSet<>();
	private final Set<BindingNode> selectedBindingNodes = new HashSet<>();

	private final BiMap<CNetRelation, DotEdge> relationToEdge;
	private final BiMap<CNetNode, DotNode> activityToNode;

	private final NodeDecorator nodeDecorator;
	private final EdgeDecorator edgeDecorator;

	private PopupDrawer popupDrawer = new PopupDrawer();

	private double bindingDelta = 8;

	CausalNetDotPanel(Dot dot, CausalNet causalNet, NodeDecorator nodeDecorator, EdgeDecorator edgeDecorator,
			BiMap<CNetRelation, DotEdge> relationToEdge, BiMap<CNetNode, DotNode> activityToNode) {
		super(createDummyDot(dot)); // cannot override the constructor of DotPanel such that we get both the Dot and the SVG
		this.causalNet = causalNet;
		this.nodeDecorator = nodeDecorator;
		this.edgeDecorator = edgeDecorator;
		this.relationToEdge = relationToEdge;
		this.activityToNode = activityToNode;
		changeDot(dot, true);
	}

	private static Dot createDummyDot(Dot dot) {
		Dot dummyDot = new Dot();
		dummyDot.setDirection(dot.getDirection());
		return dummyDot;
	}

	public void setImage(SVGDiagram image, boolean resetView) {
		if (getDot() != null) {
			bindingNodes.clear();
			bindingNodes.addAll(computeBindingNodes(getDot(), image));
			for (BindingNode node : bindingNodes) {
				node.getSelectionListeners().add(new DefaultBindingNodeSelectionListener());
			}
			for (DotNode node : getDot().getNodes()) {
				Group group = DotPanel.getSVGElementOf(image, node);
				CNetNode cnetNode = activityToNode.inverse().get(node);
				Color fillColor = nodeDecorator.getNodeFillColor(cnetNode);
				for (int i = 0; i < group.getNumChildren(); i++) {
					SVGElement child = group.getChild(i);
					if (!(child instanceof Text)) {
						DotPanel.setCSSAttributeOf(child, "fill", fillColor);
					} else {
						DotPanel.setCSSAttributeOf(child, "fill", determineFontColor(null, fillColor));
					}
				}
			}
		}
		super.setImage(image, resetView);
	}

	private static final Color determineFontColor(final Color textColor, final Color bgColor) {
		if (textColor == null) {
			double val = Math.sqrt(.299 * Math.pow(bgColor.getRed(), 2) + .587 * Math.pow(bgColor.getGreen(), 2)
					+ .114 * Math.pow(bgColor.getBlue(), 2));
			return (val < 130) ? Color.WHITE : Color.BLACK;
		} else {
			return textColor;
		}
	}

	//TODO fiddle around with the source of the events and make sure future updates to GraphViz don't break this code

	public List<Exporter> getExporters() {
		List<Exporter> exporters = super.getExporters();
		Iterator<Exporter> iterator = exporters.iterator();
		// prevent broken PDF export
		while (iterator.hasNext()) {
			if (iterator.next() instanceof ExporterPDF) {
				iterator.remove();
			}
		}
		return exporters;
	}

	@Override
	protected boolean processMouseRelease(MouseEvent e) {
		Point point = e.getPoint();
		BindingNode node = getBindingsAtPoint(point);
		if (node != null) {
			for (MouseListener mouseListener : node.getMouseListeners()) {
				mouseListener.mouseReleased(e);
			}
		}
		return super.processMouseRelease(e);
	}

	@Override
	protected boolean processMousePress(MouseEvent e) {
		Point point = e.getPoint();
		BindingNode node = getBindingsAtPoint(point);
		if (node != null) {
			for (MouseListener mouseListener : node.getMouseListeners()) {
				mouseListener.mousePressed(e);
			}
		}
		return super.processMousePress(e);
	}

	@Override
	protected boolean processMouseClick(MouseEvent e) {

		//pass clicks to elements and process selection changes, but not if the click was already caught
		boolean selectionChange = false;
		Point point = e.getPoint();
		e.setSource(this);
		BindingNode node = getBindingsAtPoint(point);
		if (node != null) {
			for (MouseListener mouseListener : node.getMouseListeners()) {
				mouseListener.mouseClicked(e);
			}
			if (SwingUtilities.isLeftMouseButton(e)) {
				selectionChange = selectionChange || bindingProcessSelection(node, e);
			}
		}

		if (SwingUtilities.isLeftMouseButton(e) && !selectionChange && !e.isControlDown()) {
			//the user did not click on anything clickable. Remove the selection.
			selectionChange = bindingRemoveSelection();
			selectionChange |= super.processMouseClick(e);
		}

		if (!selectionChange) {
			selectionChange = super.processMouseClick(e);
		}

		if (selectionChange) {
			repaint();
		}

		return selectionChange;
	}

	@Override
	protected void paintImage(Graphics2D g) {
		super.paintImage(g);
		final AffineTransform oldTransform = adjustTransform(g);

		try {

			drawNodeSegmentDecoration(g);
			drawNodeTrafficLightDecoration(g);
			drawNodeIconDecoration(g);

			for (BindingNode binding : bindingNodes) {
				binding.decorateBinding(image, g);
			}

			for (BindingNode binding : bindingNodes) {
				binding.decoratedSelectedBinding(image, g);
			}

			for (DotElement selectedElement : getSelectedElements()) {
				if (selectedElement instanceof DotEdge) {
					CNetRelation relation = relationToEdge.inverse().get(selectedElement);
					String label = edgeDecorator.getEdgeLabel(relation.getSource(), relation.getTarget());
					if (label != null && !label.isEmpty()) {
						drawLabelOnEdge((DotEdge) selectedElement, label, g);
					}
				} else if (selectedElement instanceof DotNode) {
					CNetNode node = activityToNode.inverse().get(selectedElement);
					String label = nodeDecorator.getNodeLabel(node);
					if (label != null && !label.isEmpty()) {
						drawLabelOnNode((DotNode) selectedElement, label, g);
					}
				}
			}

		} finally {
			g.setTransform(oldTransform);
		}

	}

	private void drawNodeIconDecoration(Graphics2D g) {
		Color oldColor = g.getColor();
		Stroke oldStroke = g.getStroke();
		try {
			for (DotNode node : getNodes()) {
				CNetNode cnetNode = activityToNode.inverse().get(node);
				Image icon = nodeDecorator.getNodeIcon(cnetNode);
				if (icon != null) {
					Group element = DotPanel.getSVGElementOf(image, node);

					double xPadding = 1.5;
					double yPadding = 0.1;

					double startX = element.getShape().getBounds2D().getMinX() + xPadding;
					double startY = element.getShape().getBounds2D().getMinY() + yPadding;

					g.drawImage(icon, (int) startX, (int) startY, 5, 5, null);
				}
			}
		} finally {
			g.setColor(oldColor);
			g.setStroke(oldStroke);
		}
	}

	private void drawNodeTrafficLightDecoration(Graphics2D g) {
		Color oldColor = g.getColor();
		Stroke oldStroke = g.getStroke();
		try {
			for (DotNode node : getNodes()) {
				CNetNode cnetNode = activityToNode.inverse().get(node);
				List<TrafficLightSegment> segments = nodeDecorator.getNodeTrafficLights(cnetNode);
				if (segments != null && !segments.isEmpty()) {
					Group element = DotPanel.getSVGElementOf(image, node);

					double firstSegmentXPadding = 0.25;
					double firstSegmentYPadding = 1.25;
					double segmentPadding = 1;

					double startX = element.getShape().getBounds2D().getMaxX() - firstSegmentXPadding;
					double startY = element.getShape().getBounds2D().getMinY() + firstSegmentYPadding;

					g.setColor(Color.BLACK);
					g.setStroke(new BasicStroke(0.75f));

					double lastSize = 0;
					for (TrafficLightSegment segment : segments) {
						startX += (lastSize - segment.getSize() - segmentPadding);
						g.setColor(Color.BLACK);
						g.draw(new Ellipse2D.Double(startX, startY, segment.getSize(), segment.getSize()));
						g.setColor(segment.getColor());
						g.fill(new Ellipse2D.Double(startX, startY, segment.getSize(), segment.getSize()));
					}
				}
			}
		} finally {
			g.setColor(oldColor);
			g.setStroke(oldStroke);
		}
	}

	private void drawNodeSegmentDecoration(Graphics2D g) {
		Color oldColor = g.getColor();
		Stroke oldStroke = g.getStroke();
		try {
			for (DotNode node : getNodes()) {
				CNetNode cnetNode = activityToNode.inverse().get(node);
				List<ColorSegment> segments = nodeDecorator.getNodeColorSegments(cnetNode);
				if (segments != null) {
					Group element = DotPanel.getSVGElementOf(image, node);
					Point2D.Double startPoint = new Point2D.Double(element.getShape().getBounds2D().getMinX(),
							element.getShape().getBounds2D().getMaxY());
					double barHeight = 3.0;
					g.setStroke(new BasicStroke(0.75f));
					double width = element.getShape().getBounds2D().getWidth();
					g.setColor(Color.BLACK);
					g.draw(new Rectangle2D.Double(startPoint.getX(), startPoint.getY() - barHeight, width, 3.0));

					double segmentsWidth = width - 1.0;
					double segmentStartX = startPoint.getX() + .5; // magic numbers from graphviz
					double segmentStartY = startPoint.getY() + .375 - barHeight; // magic numbers from graphviz
					for (ColorSegment segment : segments) {
						g.setColor(segment.getColor());
						g.fill(new Rectangle2D.Double(segmentStartX, segmentStartY, segment.getSize() * segmentsWidth,
								2.125));
						segmentStartX += segment.getSize() * segmentsWidth;
					}
				}
			}
		} finally {
			g.setColor(oldColor);
			g.setStroke(oldStroke);
		}
	}

	private void drawLabelOnEdge(DotEdge dotEdge, String label, Graphics2D g) {

		GeneralPath path = getPath(image, dotEdge);
		BezierPath bezierPath = convertToBezierPath(path);

		Color oldColor = g.getColor();
		Stroke oldStroke = g.getStroke();
		Font oldFont = g.getFont();

		Point2D.Double point = new Double(path.getBounds().getMaxX(), path.getBounds().getMaxY());

		if (!bezierPath.getBeziers().isEmpty()) {
			Bezier bezier = bezierPath.getBeziers().get(0);
			bezier.eval(0.5, point);

			if (label != null && !label.isEmpty()) {
				popupDrawer.drawLabelPopup(label, g, (float) point.getX(), (float) point.getY());
			}
		}

		g.setColor(oldColor);
		g.setStroke(oldStroke);
		g.setFont(oldFont);
	}

	private void drawLabelOnNode(DotNode dotNode, String label, Graphics2D g) {

		Group element = DotPanel.getSVGElementOf(image, dotNode);

		Color oldColor = g.getColor();
		Stroke oldStroke = g.getStroke();
		Font oldFont = g.getFont();

		Point2D.Double point = new Double(element.getShape().getBounds2D().getMaxX(),
				element.getShape().getBounds2D().getMaxY());

		if (label != null && !label.isEmpty()) {
			popupDrawer.drawLabelPopup(label, g, (float) point.getX(), (float) point.getY());
		}

		g.setColor(oldColor);
		g.setStroke(oldStroke);
		g.setFont(oldFont);
	}

	/**
	 * Computes how and where to place the bindings for the causal net
	 * 
	 * @param dot
	 * @param diagram
	 * @return the binding nodes
	 */
	private Set<BindingNode> computeBindingNodes(final Dot dot, SVGDiagram diagram) {
		Set<BindingNode> bindingNodes = new HashSet<>();
		computeOutputBinding(dot, diagram, bindingNodes);
		computeInputBindings(dot, diagram, bindingNodes);
		return bindingNodes;
	}

	private void computeOutputBinding(final Dot dot, SVGDiagram diagram, Set<BindingNode> bindingNodes) {
		for (CNetNode cNode : activityToNode.keySet()) {

			DotNode node = activityToNode.get(cNode);
			Rectangle2D nodeBox = getBoundingBox(diagram, node);
			double currentXDistance = Math.max(8, calcInitialDistance(nodeBox.getWidth(), 1));
			double currentYDistance = Math.max(8, calcInitialDistance(nodeBox.getHeight(), 1));

			for (PeekingIterator<CNetBinding> iterator = Iterators
					.peekingIterator(getSortedOutputBindings(cNode).iterator()); iterator.hasNext();) {
				CNetBinding binding = iterator.next();

				RectangularShape currentBoundingShape = createBindingShape(diagram, nodeBox, currentXDistance,
						currentYDistance);

				List<Connector> connectionPoints = new ArrayList<>();
				for (CNetNode target : binding.getBoundNodes()) {
					CNetRelation relation = CNetRelation.Factory.create(cNode, target);
					DotEdge edge = relationToEdge.get(relation);

					GeneralPath path = getPath(diagram, edge);
					BezierPath bezierPath = convertToBezierPath(path);
					// all in original SVG coordinates
					if ("back".equals(edge.getOption("dir"))) {
						connectionPoints.add(computePointOnShape(currentBoundingShape,
								Lists.reverse(bezierPath.getBeziers()), 1.0d, 0.0d, 0.00001d, 0.001d));
					} else {
						connectionPoints.add(computePointOnShape(currentBoundingShape, bezierPath.getBeziers(), 0.0d,
								1.0d, 0.0001d, 0.001d));
					}
				}
				bindingNodes.add(new BindingNode(cNode, binding, BindingType.OUTPUT, activityToNode.get(cNode),
						currentBoundingShape, connectionPoints, nodeDecorator));

				if (iterator.hasNext()
						&& (binding.getBoundNodes().size() > 1 || iterator.peek().getBoundNodes().size() > 1)) {
					currentXDistance += calcBindingDelta(iterator.peek());
					currentYDistance += calcBindingDelta(iterator.peek());
				}
			}
		}
	}

	private double calcBindingDelta(CNetBinding binding) {
		if (nodeDecorator.getBindingBorderStyle(binding) == BorderStyle.DOUBLE) {
			return nodeDecorator.getBindingSize(binding) * 2 * BindingNode.DOUBLE_BORDER_FACTOR;
		}
		return nodeDecorator.getBindingSize(binding) + bindingDelta;
	}

	private void computeInputBindings(final Dot dot, SVGDiagram diagram, Set<BindingNode> bindingNodes) {
		for (CNetNode cNode : activityToNode.keySet()) {

			DotNode node = activityToNode.get(cNode);
			Rectangle2D nodeBox = getBoundingBox(diagram, node);
			double currentXDistance = Math.max(25, calcInitialDistance(nodeBox.getWidth(), 1));
			double currentYDistance = Math.max(25, calcInitialDistance(nodeBox.getHeight(), 1));

			for (PeekingIterator<CNetBinding> iterator = Iterators
					.peekingIterator(getSortedInputBindings(cNode).iterator()); iterator.hasNext();) {
				CNetBinding binding = iterator.next();

				RectangularShape currentBoundingShape = createBindingShape(diagram, nodeBox, currentXDistance,
						currentYDistance);

				List<Connector> connectionPoints = new ArrayList<>();
				for (CNetNode target : binding.getBoundNodes()) {
					CNetRelation relation = CNetRelation.Factory.create(target, cNode);
					DotEdge edge = relationToEdge.get(relation);

					GeneralPath path = getPath(diagram, edge);
					BezierPath bezierPath = convertToBezierPath(path);
					// all in original SVG coordinates
					if ("back".equals(edge.getOption("dir"))) {
						connectionPoints.add(computePointOnShape(currentBoundingShape, bezierPath.getBeziers(), 0.0d,
								1.0d, 0.00001d, 0.001d));
					} else {
						connectionPoints.add(computePointOnShape(currentBoundingShape,
								Lists.reverse(bezierPath.getBeziers()), 1.0d, 0.0d, 0.00001d, 0.001d));
					}
				}
				bindingNodes.add(new BindingNode(cNode, binding, BindingType.INPUT, activityToNode.get(cNode),
						currentBoundingShape, connectionPoints, nodeDecorator));

				if (iterator.hasNext()
						&& (binding.getBoundNodes().size() > 1 || iterator.peek().getBoundNodes().size() > 1)) {
					currentXDistance += calcBindingDelta(iterator.peek());
					currentYDistance += calcBindingDelta(iterator.peek());
				}
			}
		}
	}

	protected Set<CNetBinding> getSortedOutputBindings(CNetNode cNode) {
		Set<CNetBinding> bindings = causalNet.getOutputBindings(cNode);
		if (bindings != null) {
			return sortedCopy(bindings);
		} else {
			return ImmutableSet.of();
		}
	}

	private static ImmutableSet<CNetBinding> sortedCopy(Set<CNetBinding> bindings) {
		return ImmutableSet.copyOf(Ordering.from(new Comparator<CNetBinding>() {

			public int compare(CNetBinding o1, CNetBinding o2) {
				return Ints.compare(o2.getBoundNodes().size(), o1.getBoundNodes().size());
			}
		}).immutableSortedCopy(bindings));
	}

	protected Set<CNetBinding> getSortedInputBindings(CNetNode cNode) {
		Set<CNetBinding> bindings = causalNet.getInputBindings(cNode);
		if (bindings != null) {
			return sortedCopy(bindings);
		} else {
			return ImmutableSet.of();
		}
	}

	private double calcInitialDistance(double side, double delta) {
		return ((side + delta) * Math.sqrt(2)) - side; // just the delta bigger than the side
	}

	private RectangularShape createBindingShape(SVGDiagram diagram, Rectangle2D boundingBox, double currentXDistance,
			double currentYDistance) {
		Ellipse2D.Double ellipse = new Ellipse2D.Double(boundingBox.getX() - (currentXDistance / 2d),
				boundingBox.getY() - (currentYDistance / 2d), boundingBox.getWidth() + currentXDistance,
				boundingBox.getHeight() + currentYDistance);
		return ellipse;
	}

	private Rectangle2D getBoundingBox(SVGDiagram diagram, DotNode node) {
		try {
			Group nodeGroup = DotPanel.getSVGElementOf(diagram, node);
			return nodeGroup.getBoundingBox();
		} catch (SVGException e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * Determines where the shape and the bezierPath meet each other.
	 * 
	 * @param shape
	 * @param bezierPath
	 * @param start
	 *            where to start on the bezierPath (0 = start, 1 = end)
	 * @param stop
	 *            where to stop on the bezierPath
	 * @param delta
	 *            how fine grained to search
	 * @param epsilon
	 *            when to accept a match
	 * 
	 * @return a {@link Connector} on the first {@link Point2D} that lies both
	 *         on the shape and the bezierPath
	 */
	private Connector computePointOnShape(Shape shape, List<Bezier> bezierPath, double start, double stop, double delta,
			double epsilon) {

		Double point = new Point2D.Double();
		Double startPoint = new Point2D.Double();

		Ellipse2D ellipse = (Ellipse2D) shape;
		double distance;

		for (Iterator<Bezier> iterator = bezierPath.iterator(); iterator.hasNext();) {
			Bezier bezier = iterator.next();
			double progress = start;

			startPoint = bezier.eval(progress, point);

			while ((start < stop) ? (progress <= stop) : (progress >= stop)) {
				point = bezier.eval(progress, point);

				distance = Math.abs(EllipseUtils.distanceFromEllipse(point, ellipse));
				if (distance < epsilon) {
					return new ConnectorImpl(point);
				}

				if (start < stop) {
					progress += delta;
				} else {
					progress -= delta;
				}
			}
		}

		return new ConnectorImpl(startPoint, false);

	}

	private GeneralPath getPath(SVGDiagram diagram, DotEdge edge) {
		Group element = DotPanel.getSVGElementOf(diagram, edge);
		Path path = (Path) element.getChild(1);
		Polygon poly = ((Polygon) element.getChild(2));
		GeneralPath polyPath = (GeneralPath) poly.getShape();
		GeneralPath combinedPath = new GeneralPath(path.getShape());
		if (!"back".equals(edge.getOption("dir"))) {
			combinedPath.lineTo(polyPath.getBounds2D().getCenterX(), polyPath.getBounds2D().getCenterY());
		}
		return combinedPath;
	}

	private AffineTransform adjustTransform(Graphics2D g) {
		AffineTransform oldTransform = g.getTransform();
		if (!isPaintingForPrint()) {
			g.transform(getSVGTransform(getSVG()));
		} else {
			g.transform(getSVGTransform(getSVG()));
		}
		return oldTransform;
	}

	private AffineTransform getSVGTransform(SVGDiagram svg) {
		SVGRoot root = svg.getRoot();
		AffineTransform affineTransform = root.getViewXform();
		if (root.getChild(0) instanceof Group) {
			AffineTransform secondTransform = ((Group) root.getChild(0)).getXForm();
			if (secondTransform != null) {
				affineTransform.concatenate(secondTransform);
			}
			//TODO recursively append all transforms
		}
		return affineTransform;
	}

	/**
	 * Taken from InductiveVisualMiner
	 * 
	 * @param generalPath
	 * @return a BezierPath with info about the curve
	 */
	private BezierPath convertToBezierPath(GeneralPath generalPath) {
		final List<Bezier> bezierList = new ArrayList<>();
		double curveLength = 0;

		double[] coords = new double[6];
		double sx = 0, sy = 0;

		for (PathIterator pathIt = generalPath.getPathIterator(new AffineTransform()); !pathIt.isDone(); pathIt
				.next()) {
			Bezier bezier = null;

			int segType = pathIt.currentSegment(coords);

			switch (segType) {
				case PathIterator.SEG_LINETO : {
					bezier = new Bezier(sx, sy, coords, 1);
					sx = coords[0];
					sy = coords[1];
					break;
				}
				case PathIterator.SEG_QUADTO : {
					bezier = new Bezier(sx, sy, coords, 2);
					sx = coords[2];
					sy = coords[3];
					break;
				}
				case PathIterator.SEG_CUBICTO : {
					bezier = new Bezier(sx, sy, coords, 3);
					sx = coords[4];
					sy = coords[5];
					break;
				}
				case PathIterator.SEG_MOVETO : {
					sx = coords[0];
					sy = coords[1];
					break;
				}
				case PathIterator.SEG_CLOSE :
					//Do nothing
					break;
			}

			if (bezier != null) {
				bezierList.add(bezier);
				curveLength += bezier.getLength();
			}
		}
		return new BezierPathImpl(bezierList, curveLength);
	}

	/**
	 * Process the selections based on the {@link MouseEvent} for the
	 * {@link BindingNode}. Adapted from {@link DotPanel}.
	 * 
	 * @param node
	 * @param event
	 * @return
	 */
	private boolean bindingProcessSelection(BindingNode node, MouseEvent event) {
		if (node.isSelectable()) {
			if (event.isControlDown()) {
				//only change this element
				if (selectedBindingNodes.contains(node)) {
					//deselect element
					selectedBindingNodes.remove(node);
					for (SelectionListener listener : node.getSelectionListeners()) {
						listener.deselected(node, image);
					}
				} else {
					//select element
					selectedBindingNodes.add(node);
					for (SelectionListener listener : node.getSelectionListeners()) {
						listener.selected(node, image);
					}
				}
			} else {
				if (selectedBindingNodes.contains(node)) {
					//clicked on selected element without keypress
					if (selectedBindingNodes.size() > 1) {
						//deselect all other selected elements
						Iterator<BindingNode> it = selectedBindingNodes.iterator();
						while (it.hasNext()) {
							BindingNode selectedElement = it.next();
							if (selectedElement != node) {
								for (SelectionListener listener : selectedElement.getSelectionListeners()) {
									listener.deselected(selectedElement, image);
								}
								it.remove();
							}
						}
					} else {
						//only this element was selected, deselect it
						selectedBindingNodes.remove(node);
						for (SelectionListener a : node.getSelectionListeners()) {
							a.deselected(node, image);
						}
					}
				} else {
					//clicked on not selected element without keypress
					//deselect all selected elements
					Iterator<BindingNode> it = selectedBindingNodes.iterator();
					while (it.hasNext()) {
						BindingNode selectedElement = it.next();
						if (selectedElement != node) {
							for (SelectionListener listener : selectedElement.getSelectionListeners()) {
								listener.deselected(selectedElement, image);
							}
							it.remove();
						}
					}
					//select this element
					selectedBindingNodes.add(node);
					for (SelectionListener listener : node.getSelectionListeners()) {
						listener.selected(node, image);
					}
				}
			}
			return true;
		}
		return false;
	}

	private boolean bindingRemoveSelection() {
		boolean result = false;
		for (BindingNode element : selectedBindingNodes) {
			for (SelectionListener listener : element.getSelectionListeners()) {
				listener.deselected(element, image);
			}
			result = true;
		}
		selectedBindingNodes.clear();
		return result;
	}

	private BindingNode getBindingsAtPoint(Point pointUserCoordinates) {
		if (isInImage(pointUserCoordinates)) {
			try {
				AffineTransform svgTransform = getSVGTransform(getSVG()).createInverse();
				Point2D pointImageCoordinates = transformUser2Image(pointUserCoordinates);
				pointImageCoordinates = svgTransform.transform(pointImageCoordinates, null);
				//get the elements at the clicked position
				return getNodeAtCoord(pointImageCoordinates, bindingNodes);
			} catch (NoninvertibleTransformException e) {
				return null;
			}
		}
		return null;
	}

	private BindingNode getNodeAtCoord(Point2D pointImageCoordinates, Set<BindingNode> bindingNodes) {
		for (BindingNode node : bindingNodes) {
			for (Connector connector : node.getConnectors()) {
				if (isWithinShape(pointImageCoordinates, connector.getPosition(), node.getConnectorShape())) {
					return node;
				}
			}
		}
		return null;
	}

	private boolean isWithinShape(Point2D clickedPoint, Point2D centerOfShape, Shape shape) {
		AffineTransform t = AffineTransform.getTranslateInstance(centerOfShape.getX(), centerOfShape.getY());
		return t.createTransformedShape(shape).contains(clickedPoint);
	}

	public BiMap<CNetRelation, DotEdge> getRelationToEdge() {
		return relationToEdge;
	}

	public BiMap<CNetNode, DotNode> getActivityToNode() {
		return activityToNode;
	}
}