
package org.processmining.plugins.dcnet;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.GraphicsDevice;
import java.awt.GraphicsEnvironment;
import java.awt.Image;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.IOException;
import java.text.DateFormat;
import java.util.Date;
import java.util.HashSet;
import java.util.Locale;
import java.util.Map;
import java.util.Set;

import javax.imageio.ImageIO;
import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.KeyStroke;

import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringEscapeUtils;
import org.apache.commons.lang3.StringUtils;
import org.processmining.framework.util.ui.widgets.helper.ProMUIHelper;
import org.processmining.models.cnet.CNetNode;
import org.processmining.models.cnet.CausalNet;
import org.processmining.models.dcnet.DCNet;
import org.processmining.models.dcnet.DataAwareCausalNet;
import org.processmining.plugins.graphviz.dot.Dot;
import org.processmining.plugins.graphviz.dot.Dot.GraphDirection;
import org.processmining.plugins.graphviz.dot.DotEdge;
import org.processmining.plugins.graphviz.dot.DotElement;
import org.processmining.plugins.graphviz.dot.DotNode;
import org.processmining.plugins.graphviz.visualisation.DotPanel;
import org.processmining.plugins.graphviz.visualisation.listeners.DotElementSelectionListener;

import com.google.common.collect.BiMap;
import com.google.common.collect.HashBiMap;
import com.google.common.collect.ImmutableSet;
import com.kitfox.svg.SVGDiagram;

public class DataAwareCausalNetVisualization {

	private static final String DEFAULT_FONT = "Verdana";

	private Action freezeGraph = new AbstractAction() {

		private static final long serialVersionUID = -1540697035814396137L;

		public void actionPerformed(ActionEvent e) {
			try {
				JFrame frame = new JFrame();
				frame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
				frame.setTitle("Snapshot: " + DateFormat.getDateTimeInstance().format(new Date()));
				frame.setBackground(Color.WHITE);
				frame.add(new CausalNetDotPanel(new Dot(IOUtils.toInputStream(dotPanel.getDot().toString(), "UTF-8")),
						causalNet, nodeDecorator, edgeDecorator, dotPanel.getRelationToEdge(),
						dotPanel.getActivityToNode()));
				frame.pack();
				showFrame(frame);
			} catch (IllegalArgumentException | SecurityException | IOException e1) {
				ProMUIHelper.showErrorMessage(dotPanel, e1.toString(), "Error duplicating SVG");
			}
		}

		private void showFrame(JFrame frame) {
			GraphicsEnvironment ge = GraphicsEnvironment.getLocalGraphicsEnvironment();
			GraphicsDevice[] gs = ge.getScreenDevices();
			if (gs.length > 1) {
				for (int i = 0; i < gs.length; i++) {
					if (gs[i] != dotPanel.getGraphicsConfiguration().getDevice()) {
						JFrame dummy = new JFrame(gs[i].getDefaultConfiguration());
						frame.setLocationRelativeTo(dummy);
						frame.setExtendedState(Frame.MAXIMIZED_BOTH);
						frame.setAlwaysOnTop(true);
						dummy.dispose();
						break;
					}
				}
			} else {
				Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
				frame.setLocation(screenSize.width / 6, screenSize.height / 6);
				frame.setSize(screenSize.width * 2 / 3, screenSize.height * 2 / 3);
				frame.setExtendedState(Frame.NORMAL);
			}
			frame.setVisible(true);
		}
	};

	private Action changeGraphDirection = new AbstractAction() {
		private static final long serialVersionUID = -38576326322179480L;

		public void actionPerformed(ActionEvent e) {
			GraphDirection newDirection;
			switch (dotPanel.getDot().getDirection()) {
				case bottomTop :
					newDirection = GraphDirection.leftRight;
					break;
				case leftRight :
					newDirection = GraphDirection.rightLeft;
					break;
				case rightLeft :
					newDirection = GraphDirection.topDown;
					break;
				case topDown :
					newDirection = GraphDirection.bottomTop;
					break;
				default :
					newDirection = GraphDirection.bottomTop;
					break;
			}
			direction = newDirection;
			dotPanel = createDotPanel();
			wrapperPanel.removeAll();
			wrapperPanel.add(dotPanel, BorderLayout.CENTER);
			wrapperPanel.validate();
		}
	};

	protected double ratio = 0.3d;

	private Action increaseRatio = new AbstractAction() {
		private static final long serialVersionUID = 2323619233153604381L;

		public void actionPerformed(ActionEvent e) {
			ratio = Math.min(ratio + .01, 5);
			dotPanel.getDot().setOption("ratio", getRatio());
			dotPanel.changeDot(dotPanel.getDot(), true);
		}
	};

	private Action decreaseRatio = new AbstractAction() {
		private static final long serialVersionUID = 1136135860512175161L;

		public void actionPerformed(ActionEvent e) {
			ratio = Math.max(ratio - .01, 0.05);
			dotPanel.getDot().setOption("ratio", getRatio());
			dotPanel.changeDot(dotPanel.getDot(), true);
		}
	};

	private String getRatio() {
		return String.format(Locale.ENGLISH, "%.2f", ratio);
	}

	private final CausalNet causalNet;
	private final NodeDecorator nodeDecorator;
	private final EdgeDecorator edgeDecorator;

	private final Set<CNetRelation> feedbackArcSet;

	private final JPanel wrapperPanel;
	private CausalNetDotPanel dotPanel;

	private GraphDirection direction = GraphDirection.topDown;
	private String arrowSize = "0.6";
	private boolean useFeedbackArcOptimisation = true;
	private int maxLabelLineLength = 4;

	public DataAwareCausalNetVisualization(CausalNet causalNet) {
		this(new DCNet(causalNet)); // a causal net is simply a dcnet without guards
	}

	public DataAwareCausalNetVisualization(CausalNet causalNet, NodeDecorator nodeDecorator) {
		this(new DCNet(causalNet), nodeDecorator); // a causal net is simply a dcnet without guards
	}

	public DataAwareCausalNetVisualization(CausalNet causalNet, NodeDecorator nodeDecorator,
			EdgeDecorator edgeDecorator) {
		this(new DCNet(causalNet), nodeDecorator, edgeDecorator); // a causal net is simply a dcnet without guards
	}

	public DataAwareCausalNetVisualization(final DataAwareCausalNet dataCausalNet) {
		this(dataCausalNet, new DefaultNodeDecorator(dataCausalNet), new DefaultEdgeDecorator(dataCausalNet));
	}

	public DataAwareCausalNetVisualization(DataAwareCausalNet dataCausalNet, NodeDecorator nodeDecorator) {
		this(dataCausalNet, nodeDecorator, new DefaultEdgeDecorator(dataCausalNet));
	}

	public DataAwareCausalNetVisualization(final DataAwareCausalNet dataCausalNet, NodeDecorator nodeDecorator,
			EdgeDecorator edgeDecorator) {
		super();
		this.nodeDecorator = new WrappingNodeDecorator(nodeDecorator) {

			public Image getNodeIcon(CNetNode node) {
				Image icon = super.getNodeIcon(node);
				if (icon == null && !dataCausalNet.getWrites(node).isEmpty()) {
					try {
						return ImageIO.read(getClass().getResourceAsStream("note.png"));
					} catch (IOException e) {
						return null;
					}
				}
				return icon;
			}

		};
		this.edgeDecorator = edgeDecorator;
		this.causalNet = dataCausalNet.getCausalNet();
		this.feedbackArcSet = FeedbackArcUtil.approximateFeedbackArcSet(causalNet);

		this.wrapperPanel = new JPanel(new BorderLayout());
		this.wrapperPanel.setOpaque(false);
		this.dotPanel = createDotPanel();
		this.wrapperPanel.add(dotPanel, BorderLayout.CENTER);
	}

	public JComponent getVisualization() {
		return wrapperPanel;
	}

	private CausalNetDotPanel createDotPanel() {
		Dot dot = new Dot();
		dot.setDirection(direction);
		dot.setEdgeOption("arrowsize", arrowSize);
		dot.setNodeOption("fontname", DEFAULT_FONT);
		dot.setEdgeOption("fontname", DEFAULT_FONT);

		BiMap<CNetRelation, DotEdge> relationToEdge = HashBiMap.create();
		BiMap<CNetNode, DotNode> activityToNode = HashBiMap.create();
		Set<CNetRelation> graph = createRelationSet(causalNet);
		for (CNetRelation relation : graph) {
			if (!relationToEdge.containsKey(relation)) {
				if (useFeedbackArcOptimisation) {
					relationToEdge.put(relation,
							addDependency(dot, causalNet, activityToNode, relation, feedbackArcSet));
				} else {
					relationToEdge.put(relation,
							addDependency(dot, causalNet, activityToNode, relation, ImmutableSet.<CNetRelation>of()));
				}
			}
		}

		CausalNetDotPanel dotPanel = new CausalNetDotPanel(dot, causalNet, nodeDecorator, edgeDecorator, relationToEdge,
				activityToNode);

		// Override the graph direction change
		dotPanel.getActionMap().put("changeGraphDirection", changeGraphDirection);

		addShortcut(dotPanel, "freeze in new window", "ctrl n",
				KeyStroke.getKeyStroke(KeyEvent.VK_N, InputEvent.CTRL_MASK), "newWindow", freezeGraph);

		addShortcut(dotPanel, "increase aspect ratio", "shift q",
				KeyStroke.getKeyStroke(KeyEvent.VK_Q, InputEvent.SHIFT_MASK), "increaseRatio", increaseRatio);
		addShortcut(dotPanel, "increase aspect ratio", "shift w",
				KeyStroke.getKeyStroke(KeyEvent.VK_W, InputEvent.SHIFT_MASK), "decreaseRatio", decreaseRatio);

		return dotPanel;
	}

	private void addShortcut(CausalNetDotPanel dotPanel, String description, String key, KeyStroke keyStroke,
			String actionKey, Action action) {
		dotPanel.getHelperControlsExplanations().add(description);
		dotPanel.getHelperControlsShortcuts().add(key);
		dotPanel.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).put(keyStroke, actionKey);
		dotPanel.getActionMap().put(actionKey, action);
	}

	private static Set<CNetRelation> createRelationSet(CausalNet cnet) {
		Set<CNetRelation> graph = new HashSet<>();
		for (CNetNode node : cnet.getNodes()) {
			for (CNetNode successor : cnet.getSuccessors(node)) {
				graph.add(CNetRelation.Factory.create(node, successor));
			}
			for (CNetNode predecessor : cnet.getPredecessors(node)) {
				graph.add(CNetRelation.Factory.create(predecessor, node));
			}
		}
		return graph;
	}

	private DotNode addNodeIfNeeded(Dot dot, CausalNet causalNet, final CNetNode node,
			Map<CNetNode, DotNode> activityToNode) {
		DotNode existingNode = activityToNode.get(node);
		if (existingNode == null) {
			if (node.equals(causalNet.getStartNode())) {
				DotNode startNode = createStartNode(dot);
				addMouseListener(node, startNode);
				activityToNode.put(node, startNode);
				return startNode;
			} else if (node.equals(causalNet.getEndNode())) {
				DotNode endNode = createEndNode(dot);
				addMouseListener(node, endNode);
				activityToNode.put(node, endNode);
				return endNode;
			} else {
				DotNode newNode = createNode(dot, node);
				addMouseListener(node, newNode);
				activityToNode.put(node, newNode);
				return newNode;
			}
		} else {
			return existingNode;
		}
	}

	private void addMouseListener(final CNetNode node, DotNode dotNode) {
		dotNode.setSelectable(true);
		dotNode.addSelectionListener(new DotElementSelectionListener() {

			private String defaultStrokeDashArray;
			private String defaultStrokeWidth;

			public void selected(DotElement element, SVGDiagram image) {
				defaultStrokeDashArray = DotPanel.getAttributeOf(DotPanel.getSVGElementOf(image, element),
						"stroke-dasharray");
				DotPanel.setCSSAttributeOf(image, element, "stroke-dasharray", "5,3");
				defaultStrokeWidth = DotPanel.getAttributeOf(DotPanel.getSVGElementOf(image, element), "stroke-width");
				DotPanel.setCSSAttributeOf(image, element, "stroke-width",
						String.valueOf(Integer.valueOf(defaultStrokeWidth != null ? defaultStrokeWidth : "1") * 2));
				dotPanel.repaint();
			}

			public void deselected(DotElement element, SVGDiagram image) {
				DotPanel.setCSSAttributeOf(image, element, "stroke-dasharray", defaultStrokeDashArray);
				DotPanel.setCSSAttributeOf(image, element, "stroke-width", defaultStrokeWidth);
				dotPanel.repaint();
			}
		});
		dotNode.addMouseListener(new MouseAdapter() {

			@Override
			public void mousePressed(MouseEvent e) {
				if (e.isPopupTrigger()) {
					showMenu(e);
				}
			}

			@Override
			public void mouseReleased(MouseEvent e) {
				if (e.isPopupTrigger()) {
					showMenu(e);
				}
			}

			private void showMenu(MouseEvent e) {
				JPopupMenu menu = nodeDecorator.getNodeMenu(node);
				if (menu != null && menu.getComponentCount() > 0) {
					menu.show(e.getComponent(), e.getX(), e.getY());
				}
			}

		});
	}

	protected DotNode createNode(Dot dot, CNetNode node) {
		String label = node.getLabel();
		int maxLabelLength = 101;
		if (label.length() > maxLabelLength) {
			label = StringUtils.abbreviate(label, maxLabelLength);
		}
		label = StringEscapeUtils.escapeXml10(label);
		if (label.isEmpty()) {
			label = " "; // fix to avoid the GraphViz ID from showing
		}
		label = "<<B>" + wrapLineWithSeparator(label, "<BR/>", getMaxNodeLabelLineLength()) + "</B>>";
		DotNode newNode = dot.addNode(label);
		newNode.setOption("shape", "box");
		newNode.setOption("height", ".1");
		newNode.setOption("width", ".1");
		newNode.setOption("fontsize", "12");
		return newNode;
	}

	private static String wrapLineWithSeparator(String escapeXml, String separator, int maxLineLength) {
		StringBuilder sb = new StringBuilder(escapeXml);
		int i = 0;
		while ((i = sb.indexOf(" ", i + maxLineLength)) != -1) {
			sb.replace(i, i + 1, separator);
		}
		escapeXml = sb.toString();
		return escapeXml;
	}

	protected DotNode createStartNode(Dot dot) {
		DotNode startNode = dot.addNode("");
		startNode.setOption("shape", "box");
		startNode.setOption("style", "filled, rounded");
		if (direction == GraphDirection.bottomTop || direction == GraphDirection.topDown) {
			startNode.setOption("height", ".1");
			startNode.setOption("width", ".4");
		} else {
			startNode.setOption("height", ".4");
			startNode.setOption("width", ".1");
		}
		startNode.setOption("fontsize", "8");
		return startNode;
	}

	protected DotNode createEndNode(Dot dot) {
		DotNode endNode = dot.addNode("");
		endNode.setOption("shape", "box");
		endNode.setOption("style", "filled, rounded");
		if (direction == GraphDirection.bottomTop || direction == GraphDirection.topDown) {
			endNode.setOption("height", ".1");
			endNode.setOption("width", ".4");
		} else {
			endNode.setOption("height", ".4");
			endNode.setOption("width", ".1");
		}
		endNode.setOption("fontsize", "8");
		return endNode;
	}

	protected DotEdge createEdge(Dot dot, DotNode sourceNode, DotNode targetNode) {
		return dot.addEdge(sourceNode, targetNode);
	}

	protected DotEdge createBackEdge(Dot dot, DotNode sourceNode, DotNode targetNode) {
		DotEdge edge = dot.addEdge(targetNode, sourceNode);
		edge.setOption("dir", "back");
		return edge;
	}

	private DotEdge addDependency(Dot dot, CausalNet causalNet, Map<CNetNode, DotNode> activityToNode,
			final CNetRelation relation, Set<CNetRelation> feedbackArcSet) {
		DotNode sourceNode = addNodeIfNeeded(dot, causalNet, relation.getSource(), activityToNode);
		DotNode targetNode = addNodeIfNeeded(dot, causalNet, relation.getTarget(), activityToNode);
		DotEdge edge = null;
		if (feedbackArcSet.contains(relation)) {
			edge = createBackEdge(dot, sourceNode, targetNode);
		} else {
			edge = createEdge(dot, sourceNode, targetNode);
		}
		edge.setOption("color", convertColor(edgeDecorator.getEdgeColor(relation.getSource(), relation.getTarget())));
		edge.setOption("penwidth",
				String.format(Locale.ENGLISH, "%.2f", edgeDecorator.getEdgeWidth(relation.getSource(), relation.getTarget())));
		edge.setOption("style", convertEdgeStyle(relation));
		edge.addMouseListener(new MouseAdapter() {

			@Override
			public void mousePressed(MouseEvent e) {
				if (e.isPopupTrigger()) {
					showMenu(e);
				}
			}

			@Override
			public void mouseReleased(MouseEvent e) {
				if (e.isPopupTrigger()) {
					showMenu(e);
				}
			}

			private void showMenu(MouseEvent e) {
				JPopupMenu menu = edgeDecorator.getEdgeMenu(relation.getSource(), relation.getTarget());
				if (menu != null && menu.getComponentCount() > 0) {
					menu.show(e.getComponent(), e.getX(), e.getY());
				}
			}

		});
		edge.addSelectionListener(new DotElementSelectionListener() {

			public void selected(DotElement element, SVGDiagram image) {
				DotPanel.setCSSAttributeOf(image, element, "stroke-dasharray", "5,3");
				dotPanel.repaint();
			}

			public void deselected(DotElement element, SVGDiagram image) {
				DotPanel.setCSSAttributeOf(image, element, "stroke-dasharray", null);
				dotPanel.repaint();
			}
		});
		return edge;
	}

	private String convertEdgeStyle(CNetRelation relation) {
		switch (edgeDecorator.getEdgeStyle(relation.getSource(), relation.getTarget())) {
			case DASHED :
				return "dashed";
			case DOTTED :
				return "dotted";
			case BOLD :
				return "bold";
			//$FALL-THROUGH$
			case SOLID :
			default :
				return "solid";
		}
	}

	private static String convertColor(Color color) {
		return String.format(Locale.ENGLISH,"#%02X%02X%02X", color.getRed(), color.getGreen(), color.getBlue());
	}

	public String getArrowSize() {
		return arrowSize;
	}

	public void setArrowSize(String arrowSize) {
		this.arrowSize = arrowSize;
	}

	public int getMaxNodeLabelLineLength() {
		return maxLabelLineLength;
	}

	public void setMaxNodeLabelLineLength(int maxLabelLineLength) {
		this.maxLabelLineLength = maxLabelLineLength;
	}

	public boolean isUseFeedbackArcOptimisation() {
		return useFeedbackArcOptimisation;
	}

	public void setUseFeedbackArcOptimisation(boolean useFeedbackArcOptimisation) {
		this.useFeedbackArcOptimisation = useFeedbackArcOptimisation;
	}

}